import os
import json

with open(os.environ.get('PYTHON_RUN_PATH')+"/completion", "r") as h:
    completion = h.read()

function = None
if os.path.exists(os.environ.get('PYTHON_RUN_PATH')+"/function"):
    with open(os.environ.get('PYTHON_RUN_PATH')+"/function", "r") as h:
        function = h.read()

try:
    context = onNext(ChainContext(message=completion, function=function))
    if context.do.overwriteMessage is not None:
        print("<overwriteMessage>" + context.do.overwriteMessage + "</overwriteMessage>")
    if context.do.message is not None:
        print("<message>" + context.do.message + "</message>")
    elif context.do.function is not None:
        print("<function>" + json.dumps(context.do.function.model_dump()) + "</function>")
    elif context.do.finalAnswer is not None:
        print("<finalAnswer>" + context.do.finalAnswer + "</finalAnswer>")
except Exception as e:
    print("<error>"+str(e)+"</error>")


import json
from runnable_hub import RunnableWorker, RunnableContext, RunnableStatus, RunnableOutputLoads
from .request.apiRequest import ApiRequest, ApiHttpMethod
from .response import ApiResponse
import aiohttp

class Worker(RunnableWorker):

    runnableCode = "API"
    Request = ApiRequest
    Response = ApiResponse

    def __init__(self):
        pass

    @staticmethod
    def outputs(raw, loadType: RunnableOutputLoads):
        if loadType == RunnableOutputLoads.JSON:
            return json.loads(raw)
        else:
            return raw

    async def onNext(self, context: RunnableContext[ApiRequest, ApiResponse]) -> RunnableContext:
        
        if context.request.method == ApiHttpMethod.GET:
            async with aiohttp.ClientSession() as session:
                async with session.get(context.request.url, 
                                       headers=context.request.headers, params=context.request.params) as response:
                    result = await response.text()
                    context.response = ApiResponse(
                        result=result, outputs=self.outputs(result, context.request.outputLoads),
                        statusCode=response.status)
                    context.status = RunnableStatus.SUCCESS
                    return context
        elif context.request.method == ApiHttpMethod.POST:
            async with aiohttp.ClientSession() as session:
                async with session.post(context.request.url, 
                                        headers=context.request.headers, params=context.request.params, json=context.request.payloads) as response:
                    result = await response.text()
                    context.response = ApiResponse(
                        result=result, outputs=self.outputs(result, context.request.outputLoads),
                        statusCode=response.status)
                    context.status = RunnableStatus.SUCCESS
                    return context
        else:
            context.status = RunnableStatus.ERROR
            context.errorMessage = "method not support"
            return context
    








from pydantic import BaseModel
from typing import List, Dict, Optional

class ProcessStep(BaseModel):
    id: str
    runnableCode: Optional[str] = None
    request: Optional[Dict] = None

    # Common node customization fields
    shell: Optional[str] = None
    api: Optional[Dict|str] = None
    jinja: Optional[Dict] = None
    python: Optional[str] = None



from pydantic import BaseModel
from runnable_hub import RunnableValueDefine, RunnableValueDefineType
from typing import Dict, List, Optional
from .agentChainTemplate import AgentChainTemplate
from ...chainWorker.request.chainFunction import ChainFunction
from ...llmWorker.request.llmSetting import LlmSetting


class AgentDefine(BaseModel):
    agentCode: str
    agentVersion: str
    prerun: Optional[Dict] = None
    postrun: Optional[Dict] = None
    inputDefine: List[RunnableValueDefine] = [RunnableValueDefine(name="prompt", type=RunnableValueDefineType.STRING, required=True)]
    outputDefine: List[RunnableValueDefine] = []
    instanceDefine: List[RunnableValueDefine] = []

    chainFunctions: List[ChainFunction]
    instruction: str
    
    llm: Optional[LlmSetting]
    llmCode: Optional[str]

    chainTemplate: Optional[AgentChainTemplate]
    chainTemplateCode: Optional[str]


from pydantic import BaseModel
from enum import Enum
from typing import Optional, Dict, List


class RunnableOutputLoads(Enum):
    JSON = "JSON"
    TEXT = "TEXT"
    YAML = "YAML"

class RunnableValueDefineType(Enum):
    NUMBER = "NUMBER"
    STRING = "STRING"
    BOOLEAN = "BOOLEAN"

class RunnableValueDefine(BaseModel):
    name: str
    type: RunnableValueDefineType
    required: Optional[bool] = False
    defaultValue: Optional[Dict|List|str|bool|float] = None



from pydantic import BaseModel
from runnable_hub import RunnableOutputLoads
from .toolType import ToolType
from .toolParamSpec import ToolParamSpec
from typing import Dict, List, Optional


class ToolDefine(BaseModel):
    toolCode: str
    toolVersion: str
    toolType: ToolType
    setting: Dict                            # 工具配置
    inputSpec: List[ToolParamSpec] = []      # 输入变量结构
    outputSpec: List[ToolParamSpec] = []     # 输出变量结构
    outputsLoads: RunnableOutputLoads = RunnableOutputLoads.TEXT
    outputTemplate: Optional[str] = None     # 输出模板




from runnable_hub import RunnableResponse
from typing import Dict, Optional, Any

class ProcessResponse(RunnableResponse):
    runnableCode: str = "PROCESS"
    outputs: Optional[Dict|str] = None
    



from pydantic import BaseModel
from typing import List, Dict, Optional, Generic, TypeVar
from datetime import datetime
from enum import Enum

T = TypeVar('T')
R = TypeVar("R")

class RunnableStatus(Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"

class RunnableRequest(BaseModel):
    runnableCode: str

class RunnableResponse(BaseModel):
    runnableCode: str

class RunnablePromise(BaseModel):
    resolve: Dict[str, Dict] = {}
    result: Dict[str, Dict] = {}
    reject: Dict[str, str] = {}

class RunnableContext(BaseModel, Generic[T,R]):
    request: T
    response: Optional[R] = None
    promise: RunnablePromise = RunnablePromise()
    executeId: str
    createTime: datetime
    startTime: Optional[datetime] = None
    endTime: Optional[datetime] = None
    status: RunnableStatus
    errorMessage: Optional[str] = None
    callDepth: int = 0
    data: Dict = {}
    runnableCode: str
    parentRunnableCode: Optional[str] = None
    parentExecuteId: Optional[str] = None
    name: Optional[str] = None
    storePath: str




from datetime import datetime
import asyncio
import uuid
from abc import ABC, abstractmethod
from .context import RunnableRequest, RunnableContext, RunnableStatus
from .store import RunnableFileStore
from typing import Dict
import os
import traceback

class RunnableWorker(ABC):
    runnableCode = None
    Request = None
    Response = None
    @abstractmethod
    async def onNext(self, context: RunnableContext) -> RunnableContext:
        pass


class RunnableHub():

    def __init__(self, store: RunnableFileStore):
        self.workers = {}
        self.requests = {}
        self.responses = {}
        self.store = store


    @staticmethod
    def shortExecuteId(executeId):
        return executeId.split("-")[0]

    def registerWorker(self, worker: RunnableWorker):
        self.workers[worker.runnableCode] = RunnableWorkerDispatch(
            worker=worker,
            queue=asyncio.Queue(),
            store=self.store,
            hub=self
        )
        self.requests[worker.runnableCode] = worker.Request
        self.responses[worker.runnableCode] = worker.Response

    def readExecuteContext(self, storePath:str, runnableCode: str) -> RunnableContext:
        return RunnableContext[self.requests[runnableCode], self.responses[runnableCode]].model_validate_json(
            self.store.readFile(f"{storePath}/context.json"))

    async def executeStart(self, request: RunnableRequest, parentContext: RunnableContext|None = None, name: str|None = None) -> RunnableContext:
        executeId=str(uuid.uuid4())
        if parentContext is None:
            storePath = f"execute/{executeId}"
            parentExecuteId = None
            parentRunnableCode = None
            callDepth = 0
        else:
            storePath = f"{parentContext.storePath}/{self.shortExecuteId(executeId)}"
            parentExecuteId = parentContext.executeId
            parentRunnableCode = parentContext.runnableCode
            callDepth = parentContext.callDepth + 1

        newContext = RunnableContext(
            executeId=executeId,
            runnableCode=request.runnableCode,
            request=request, 
            response=None, 
            name=name,
            callDepth=callDepth,
            storePath=storePath,
            parentExecuteId=parentExecuteId,
            parentRunnableCode=parentRunnableCode,
            createTime=datetime.now(), 
            status=RunnableStatus.PENDING)
        
        self.store.saveFile(f"{newContext.storePath}/context.json", newContext.model_dump_json())
        await self.workers[newContext.runnableCode].queue.put(f"{newContext.storePath}/context.json|")
        return newContext
    
    async def parentExecuteNext(self, context: RunnableContext):
        if context.parentExecuteId is None or context.parentRunnableCode is None or context.name is None:
            return
        parentStorePath = os.path.dirname(context.storePath)
        await self.workers[context.parentRunnableCode].queue.put(f"{parentStorePath}/context.json|{context.runnableCode}#{context.name}={self.shortExecuteId(context.executeId)}")

    async def executeWait(self, context: RunnableContext) -> RunnableContext:
        while context.status not in [RunnableStatus.ERROR, RunnableStatus.SUCCESS]:
            await asyncio.sleep(1)
            context = self.readExecuteContext(context.storePath, context.runnableCode)
        return context

class RunnableWorkerDispatch():
    worker: RunnableWorker
    queue: asyncio.Queue
    store: RunnableFileStore
    hub: RunnableHub

    def __init__(self, worker: RunnableWorker, queue: asyncio.Queue, store: RunnableFileStore, hub: RunnableHub):
        self.worker = worker
        self.queue = queue
        self.store = store
        self.hub = hub
        asyncio.create_task(self.run())

    async def run(self):
        while True:
            print(f"RunnableWorkerDispatch {self.worker.runnableCode} wait")
            message = await self.queue.get()
            contextFile, callbacks = message.split("|", 1)
            if callbacks != "":
                print(f"RunnableWorkerDispatch {self.worker.runnableCode} get message {contextFile} with callback {callbacks}")
            else:
                print(f"RunnableWorkerDispatch {self.worker.runnableCode} get message {contextFile}")

            context = RunnableContext[self.worker.Request, self.worker.Response].model_validate_json(
                self.store.readFile(contextFile))

            if callbacks != "":
                for callback in callbacks.split(","):
                    # runnableCode#name=executeShortId
                    runnableCodeAndName, executeShortId = callback.split('=', 1)
                    callbackRunnableCode, name = runnableCodeAndName.split('#', 1)
                    callbackContext = self.hub.readExecuteContext(f"{context.storePath}/{executeShortId}", callbackRunnableCode)
                    if callbackContext.status == RunnableStatus.SUCCESS:
                        context.promise.result[name] = callbackContext.response.model_dump() if callbackContext.response is not None else {}
                    elif callbackContext.status == RunnableStatus.ERROR:
                        context.promise.reject[name] = callbackContext.errorMessage if callbackContext.errorMessage is not None else ""
                    else:
                        print(f"context {context.storePath}/{executeShortId} not in finish status")


            if context.status == RunnableStatus.PENDING:
                context.startTime = datetime.now()
                context.status = RunnableStatus.RUNNING
            elif context.status in [RunnableStatus.ERROR, RunnableStatus.SUCCESS]:
                print(f"context {context.executeId} has been finished, status={context.status}")
                continue
        
            try:
                context = await self.worker.onNext(context)
            except Exception as e:
                context.status = RunnableStatus.ERROR
                context.errorMessage = f"Error: {str(e)}\nStack Trace:\n{traceback.format_exc()}"

            # 执行器读取完成后将此结果置空
            context.promise.result = {}

            if context.status in [RunnableStatus.ERROR, RunnableStatus.SUCCESS]:
                context.endTime = datetime.now()
                await self.hub.parentExecuteNext(context)
            elif context.status == RunnableStatus.RUNNING:
                todo = context.promise.resolve
                while todo:
                    name, req = todo.popitem()
                    print(f"get todo {name} {req}")
                    await self.hub.executeStart(self.hub.requests[req['runnableCode']](**req), context, name)

            print(context)
            self.store.saveFile(contextFile, context.model_dump_json())




from runnable_hub import RunnableRequest, RunnableOutputLoads
from .llmMessage import LlmMessage
from .llmSetting import LlmSetting
from typing import Dict, List

class LlmRequest(RunnableRequest):
    runnableCode: str = "LLM"
    setting: LlmSetting
    systemPrompt: str
    userPrompt: str
    history: List[LlmMessage] = []


import json
from re import L
from runnable_hub import RunnableWorker, RunnableContext, RunnableStatus, RunnableOutputLoads
from .request.llmRequest import LlmRequest
from .response.llmResponse import LlmResponse, LlmMessage, LlmUsage
import aiohttp

class Worker(RunnableWorker):

    runnableCode = "LLM"
    Request = LlmRequest
    Response = LlmResponse

    def __init__(self):
        pass

    async def onNext(self, context: RunnableContext[LlmRequest, LlmResponse]) -> RunnableContext:
        headers = {
            "Authorization": f"Bearer {context.request.setting.secretKey}",
            "Content-Type": "application/json"
        }
        payload = {
            "model": context.request.setting.model,
            "messages":[{
                "role": "system",
                "content": context.request.systemPrompt
            }]
        }
        for message in context.request.history:
            payload["messages"].append({
                "role": message.role,
                "content": message.content
            })
        payload["messages"].append({
            "role": "user",
            "content": context.request.userPrompt
        })

        async with aiohttp.ClientSession() as session:
            async with session.post(context.request.setting.endpoint, 
                                    headers=headers, json=payload) as response:
                result = await response.json()
                messages = [LlmMessage(
                    role="user",
                    content=context.request.userPrompt,
                ),LlmMessage(
                    role=result["choices"][0]["message"]["role"],
                    content=result["choices"][0]["message"]["content"]
                )]
                usage = LlmUsage(
                    promptTokens=result["usage"]["prompt_tokens"],
                    completionTokens=result["usage"]["completion_tokens"],
                    totalTokens=result["usage"]["total_tokens"]
                )
                context.response = LlmResponse(
                    messages=messages,
                    usage=usage,
                    content=result["choices"][0]["message"]["content"],
                )
                context.status = RunnableStatus.SUCCESS
                return context


  







from runnable_hub import RunnableWorker, RunnableContext, RunnableStatus, RunnableOutputLoads
from .request.jinjaRequest import JinjaRequest
from .response import JinjaResponse
from jinja2 import Environment
import json

class Worker(RunnableWorker):

    runnableCode = "JINJA"
    Request = JinjaRequest
    Response = JinjaResponse

    def __init__(self):
        self.jinjaEnv = Environment()


    async def onNext(self, context: RunnableContext[JinjaRequest, JinjaResponse]) -> RunnableContext:
        
        renderData = context.request.data

        rawResult = self.jinjaEnv.from_string(context.request.template).render(**renderData)
        if context.request.outputLoads == RunnableOutputLoads.TEXT:
            context.response = JinjaResponse(result=rawResult, outputs=rawResult)
            context.status = RunnableStatus.SUCCESS
        elif context.request.outputLoads == RunnableOutputLoads.JSON:
            context.response = JinjaResponse(result=rawResult, outputs=json.loads(rawResult))
            context.status = RunnableStatus.SUCCESS
        else:
            context.status = RunnableStatus.ERROR
            context.errorMessage = "resultFormat not support"
        
        return context








from typing import Dict, Optional, List
from runnable_hub import RunnableResponse

class AgentResponse(RunnableResponse):
    runnableCode: str = "AGENT"
    outputs: Dict|List|str
    errorMessage: Optional[str] = None
    success: bool




from pydantic import BaseModel
from typing import Optional, Dict, List

from traitlets import default
from .toolParamSpecType import ToolParamSpecType

class ToolApiSetting(BaseModel):
    payload: Optional[Dict] = None
    headers: Optional[Dict] = None
    params: Optional[Dict] = None
    url: str
    method: str



from runnable_hub import RunnableWorker, RunnableContext, RunnableStatus
from .request.shellRequest import ShellRequest
from .response import ShellResponse

import asyncio
import tempfile
import shutil
import os

class Worker(RunnableWorker):

    runnableCode = "SHELL"
    Request = ShellRequest
    Response = ShellResponse

    shellBin = "/bin/bash"

    def __init__(self, storePath = "/tmp/shell"):
        self.storePath = storePath
        if not os.path.exists(self.storePath):
            os.makedirs(self.storePath)

    @staticmethod
    async def run_command(command, cwd=None, env=None):
        process = await asyncio.create_subprocess_exec(
            *command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=cwd,
            env=env
        )
        stdout, stderr = await process.communicate()
        return process.returncode, stdout.decode(), stderr.decode()

    async def onNext(self, context: RunnableContext[ShellRequest, ShellResponse]) -> RunnableContext:
        
        # create a temporary path to store script
        temporary_path = os.path.join(self.storePath, context.executeId)
        if not os.path.exists(temporary_path):
            os.makedirs(temporary_path)

        temp_file = f"{temporary_path}/run.sh"
        with open(temp_file, "w") as h:
            h.write(context.request.run)
        
        try:
            returncode, stdout, stderr = await self.run_command(
                [self.shellBin, temp_file], cwd=temporary_path, env={"SHELL_RUN_PATH": temporary_path})
            outputs = {}
            if context.request.outputs is not None:
                for key, fileName in context.request.outputs.items():
                    with open(f"{temporary_path}/{fileName}", "r") as h:
                        outputs[key] = h.read().strip()

            context.response = ShellResponse(returncode=returncode, stdout=stdout, stderr=stderr, outputs=outputs)
            context.status = RunnableStatus.SUCCESS
                
        finally:
            shutil.rmtree(temporary_path)

        return context







from runnable_hub import RunnableRequest, RunnableOutputLoads
from typing import Dict, Optional, List

class JinjaRequest(RunnableRequest):
    runnableCode: str = "JINJA"
    data: Dict[str, Dict|List|str|float|int|bool] = {}
    template: str
    outputLoads: RunnableOutputLoads = RunnableOutputLoads.TEXT




from runnable_hub import RunnableResponse
from typing import Dict, Optional, List

class ChainResponse(RunnableResponse):
    runnableCode: str = "CHAIN"
    finalAnswer: str
    history: List[Dict] = []




from datetime import datetime
from typing import Dict, Any, List
from runnable_hub import RunnableWorker, RunnableContext, RunnableStatus
from .request.processRequest import ProcessRequest
from .request.processStep import ProcessStep
from .response import ProcessResponse
from jinja2 import Environment
from uuid import uuid4
import json

class OutputRender:
    def __init__(self, pathValueTemplate):
        self.outputValueMap = {}
        self.pathValueTemplate = pathValueTemplate
    def __getattr__(self, name):
        self.outputValueMap[name] = str(uuid4())[:8]+".txt"
        return {"path": self.pathValueTemplate.format(outputFileName=self.outputValueMap[name])}

def save_filter(value, target_dict):
    target_dict["save"] = value
    return value

class Worker(RunnableWorker):

    runnableCode = "PROCESS"
    Request = ProcessRequest
    Response = ProcessResponse

    def __init__(self):
        self.jinjaNewEnv = Environment(
            variable_start_string="${{",  # 使用 ${{ 作为变量开始标记
            variable_end_string="}}",     # 使用 }} 作为变量结束标记
        )
        self.jinjaNewEnv.filters["save"] = save_filter

    def complexRender(self, data: Dict, target):
        if isinstance(target, dict):
            return {self.complexRender(data, key): self.complexRender(data, value) for key, value in target.items()}
        elif isinstance(target, list):
            return [self.complexRender(data, value) for value in target]
        elif isinstance(target, str):
            target = target.strip()
            # 如果整个字符串只有 ${{ ... }} 则直接取里面的内容，保留变量类型
            if target.startswith("${{") and target.endswith("}}"):
                fetchValue = {}
                self.jinjaNewEnv.from_string(target[:-2] + "|save(fetchValue)}}").render(fetchValue=fetchValue, **data)
                return fetchValue["save"]
            else:
                return self.jinjaNewEnv.from_string(target).render(**data)
        else:
            return target

    def make_worker_request(self, step, stepOutputs) -> Dict:

        # stepRender:Any = self.complexRender({"outputs": outputs, "steps": stepOutputs}, step)
        if step.get("shell") is not None:
            outputs = OutputRender('$SHELL_RUN_PATH/{outputFileName}')
            stepRender:Any = self.complexRender({"outputs": outputs, "steps": stepOutputs}, step)
            stepRender["request"] = {
                "runnableCode": "SHELL",
                "run": stepRender["shell"],
                "outputs": outputs.outputValueMap,
            }
        elif step.get("api") is not None:
            stepRender:Any = self.complexRender({"steps": stepOutputs}, step)
            stepRender["request"] = stepRender["api"]
            stepRender["request"]["runnableCode"] = "API"
        elif step.get("jinja") is not None:
            stepRender:Any = self.complexRender({"steps": stepOutputs}, step)
            stepRender["request"] = stepRender["jinja"]
            stepRender["request"]["runnableCode"] = "JINJA"
        elif step.get("python") is not None:
            outputs = OutputRender("os.environ.get('PYTHON_RUN_PATH')+'/{outputFileName}'")
            stepRender:Any = self.complexRender({"outputs": outputs, "steps": stepOutputs}, step)
            stepRender["request"] = {
                "runnableCode": "PYTHON",
                "run": stepRender["python"],
                "outputs": outputs.outputValueMap,
            }
            stepRender["request"]["runnableCode"] = "PYTHON"
        else:
            stepRender:Any = self.complexRender({"steps": stepOutputs}, step)
            stepRender["request"]["runnableCode"] = stepRender["runnableCode"]
        return stepRender


    async def onNext(self, context: RunnableContext[ProcessRequest, ProcessResponse]) -> RunnableContext:
        
        if context.data.get("runtime") is None:
            context.data["runtime"] = {}
            for jobId, job in context.request.jobs.items():
                if len(job.steps) == 0:
                    raise RuntimeError(f"jobId {jobId} steps is empty")
                context.data["runtime"][jobId] = {
                    "needs": job.needs[:],
                    "steps": [step.model_dump() for step in job.steps],
                    "stepOutputs": {},
                    "errors": {},
                    "currentStepId": None,
                    "startTime": datetime.now(),
                    "outputs": job.outputs,
                }
                if len(job.needs) == 0:
                    context.data["runtime"][jobId]["jobStatus"] = "RUNNING"
                    context.data["runtime"][jobId]["stepStatus"] = "RUNNING"
                else:
                    context.data["runtime"][jobId]["jobStatus"] = "PENDING"
                    context.data["runtime"][jobId]["stepStatus"] = "PENDING"

        for jobId,job in context.data["runtime"].items():
            if job["jobStatus"] != "RUNNING":
                continue

            if context.promise.result.get(jobId) is not None:
                job["stepOutputs"][job["currentStepId"]] = context.promise.result[jobId]
                job["currentStepId"] = None
                job["stepStatus"] = "SUCESS"
            
            elif context.promise.reject.get(jobId) is not None:
                job["errors"][job["currentStepId"]] = context.promise.reject[jobId]
                job["currentStepId"] = None
                job["stepStatus"] = "ERROR"

            if job["currentStepId"] is not None: # job not finish
                continue

            if job["stepStatus"] == "ERROR":
                job["jobStatus"] = "ERROR"
                job["endTime"] = datetime.now()
                for checkJob in context.data["runtime"].values():
                    checkJob["jobStatus"] = "ERROR"
                    checkJob["endTime"] = datetime.now()
            elif len(job["steps"]) > 0:
                step = self.make_worker_request(job["steps"].pop(0), job["stepOutputs"])
                job["currentStepId"] = step["id"]
                context.promise.resolve[jobId] = step["request"]
            else:
                job["jobStatus"] = "SUCCESS"
                job["endTime"] = datetime.now()
                if job["outputs"] is not None:
                    job["outputs"] = self.complexRender({"steps": job["stepOutputs"]}, job["outputs"])
                for checkJob in context.data["runtime"].values():
                    if jobId in checkJob["needs"]:
                        checkJob["needs"].remove(jobId)
                        if len(checkJob["needs"]) == 0 and checkJob["jobStatus"] == "PENDING":
                            checkJob["jobStatus"] = "RUNNING"
                            checkJob["stepStatus"] = "RUNNING"

        allStatus = set([r["jobStatus"] for r in context.data["runtime"].values()])
        if allStatus == set(["SUCCESS"]):
            context.status = RunnableStatus.SUCCESS
            if context.request.outputs is not None:
                jobOutputs = {jobId: {"outputs":job["outputs"]} for jobId, job in context.data["runtime"].items()}
                processOutputs = self.complexRender({"jobs": jobOutputs}, context.request.outputs)
                context.response = ProcessResponse(outputs=processOutputs) # type: ignore

        elif allStatus == set(["ERROR"]) or allStatus == set(["ERROR","SUCCESS"]):
            context.status = RunnableStatus.ERROR

        return context






from runnable_hub import RunnableRequest
from typing import Dict, Optional
from .processJob import ProcessJob

class ProcessRequest(RunnableRequest):
    runnableCode: str = "PROCESS"
    jobs: Dict[str, ProcessJob]
    outputs: Optional[Dict|str] = None



from pydantic import BaseModel
from typing import Optional, Dict, List

from .toolParamSpecType import ToolParamSpecType

class ToolParamSpec(BaseModel):
    name: str
    type: ToolParamSpecType
    required: Optional[bool] = False
    defaultValue: Optional[Dict|List|str|bool|float] = None


from .hub import RunnableHub, RunnableWorker
from .context import RunnableContext, RunnableRequest, RunnableStatus, RunnableResponse
from .common import RunnableOutputLoads, RunnableValueDefine, RunnableValueDefineType



from runnable_hub import RunnableRequest
from typing import Dict, Optional

class PythonRequest(RunnableRequest):
    runnableCode: str = "PYTHON"
    data: Dict[str, str] = {}
    run: str
    outputs: Optional[Dict[str, str]] = None          # { "variable name": "output file path" }




from runnable_hub import RunnableRequest
from typing import Dict, Optional, List
from .chainFunction import ChainFunction
from ...llmWorker.request.llmSetting import LlmSetting

class ChainRequest(RunnableRequest):
    runnableCode: str = "CHAIN"
    functions: List[ChainFunction] = []
    data: Dict[str, Dict|List|str|float|int|bool] = {}
    llm: LlmSetting
    systemPrompt: str
    userPrompt: str
    onNext: str
    # chainInterpreter: str





from pydantic import BaseModel
from typing import List, Dict, Optional
from .processStep import ProcessStep

class ProcessJob(BaseModel):
    steps: List[ProcessStep]
    needs: List[str] = []
    outputs: Optional[Dict|str] = None



from runnable_hub import RunnableResponse
from typing import Dict, List, Optional

class ApiResponse(RunnableResponse):
    runnableCode: str = "API"
    result: str
    statusCode: int
    outputs: Optional[Dict|List|str]




from runnable_hub import RunnableResponse
from typing import Dict, Optional, List

class JinjaResponse(RunnableResponse):
    runnableCode: str = "JINJA"
    result: str
    outputs: Optional[Dict|List|str]





from runnable_hub import RunnableResponse
from typing import Dict, Optional

class PythonResponse(RunnableResponse):
    runnableCode: str = "PYTHON"
    stdout: str
    stderr: str
    returncode: int | None
    outputs: Dict[str, str] = {}



from runnable_hub import RunnableRequest
from typing import Dict

class AgentRequest(RunnableRequest):
    runnableCode: str = "AGENT"
    agentCode: str
    agentVersion: str
    inputs: Dict
    prompt: str


from json import tool
from pydantic import BaseModel
from typing import Dict, Optional, List
from enum import Enum
from runnable_hub import RunnableValueDefine

class ChainFunctionType(str, Enum):
    TOOL = "TOOL"
    AGENT = "AGENT"

class ChainFunction(BaseModel):
    type: ChainFunctionType
    name: str
    version: str
    inputDefine: List[RunnableValueDefine] = []
    outputDefine: List[RunnableValueDefine] = []
    presetInputs: Dict = {}
    description: Optional[str] = None





from typing import Dict
from runnable_hub import RunnableWorker, RunnableContext, RunnableStatus
from .request.chainRequest import ChainRequest, ChainFunction
from .request.chainFunction import ChainFunctionType
from .response import ChainResponse
import sys
from jinja2 import Environment

import os
import re
import json


# enter   -->     llmEnd        <----   llmStart
#                    |                      ^ 
#                    |                      |
#                    v                      |    
#               functionCallStart  --->  functionCallEnd  --->  finish
            

class Worker(RunnableWorker):

    runnableCode = "CHAIN"
    Request = ChainRequest
    Response = ChainResponse

    pythonBin = sys.executable

    def __init__(self, storePath = "/tmp/python"):
        self_dir = os.path.dirname(__file__)
        self.jinjaEnv = Environment()
        with(open(self_dir + "/next.py", "r")) as h:
            self.nextPostScript = h.read()

    @staticmethod
    def fnInputMerge(fn: ChainFunction, inputs: Dict|str) -> Dict:
        finalInputs = fn.presetInputs.copy()
        if isinstance(inputs, str):
            finalInputs.update({
                fn.inputDefine[0].name: inputs
            })
        else:
            finalInputs.update(inputs)
        return finalInputs


    # @staticmethod
    # async def run_python(command, cwd=None, env=None):
    #     process = await asyncio.create_subprocess_exec(
    #         *command,
    #         stdout=asyncio.subprocess.PIPE,
    #         stderr=asyncio.subprocess.PIPE,
    #         cwd=cwd,
    #         env=env
    #     )
    #     stdout, stderr = await process.communicate()
    #     return process.returncode, stdout.decode(), stderr.decode()


    async def onNext(self, context: RunnableContext[ChainRequest, ChainResponse]) -> RunnableContext:

        if context.data.get("runtime") is None:
            # todo merge function call defines
            renderData = context.request.data

            renderData["tool_info"] = ""
            renderData["function_info"] = ""
            renderData["agent_info"] = ""
            for fn in context.request.functions:
                args = ", ".join([f"{arg.name}: {arg.type.value}" for arg in fn.inputDefine])
                info_block = f"{fn.name}: {fn.name}({args}) - {fn.description}\n"
                renderData["function_info"] += info_block
                if fn.type == ChainFunctionType.TOOL:
                    renderData["tool_info"] += info_block
                elif fn.type == ChainFunctionType.AGENT:
                    renderData["agent_info"] += info_block

            systemPrompt = self.jinjaEnv.from_string(context.request.systemPrompt).render(**renderData)
            userPrompt = self.jinjaEnv.from_string(context.request.userPrompt).render(**renderData)

            context.promise.resolve["llm"] = {
                "runnableCode": "LLM",
                "setting": context.request.llm.model_dump(),
                "systemPrompt": systemPrompt,
                "userPrompt": userPrompt,
            }

            context.data["runtime"] = {
                "history": [],
                "systemPrompt": systemPrompt,
                "nextStep": "llmEnd",
                "lastCompletion": None,
            }
            return context
        
        elif context.data["runtime"]["nextStep"] == "llmEnd":
            
            if context.promise.result["llm"] is None:
                raise ValueError("LLM response is missing")
            
            context.promise.resolve["processMessage"] = {
                "runnableCode": "PYTHON",
                "data": {
                    "completion": context.promise.result["llm"]["content"],
                },
                "run": context.request.onNext + "\n" + self.nextPostScript,
            }
            context.data["runtime"]["nextStep"] = "functionCallStart"
            context.data["runtime"]["lastCompletion"] = context.promise.result["llm"]["content"]
            context.data["runtime"]["history"] += context.promise.result["llm"]["messages"]
            return context

        elif context.data["runtime"]["nextStep"] == "functionCallStart":

            if context.promise.result["processMessage"] is None:
                raise ValueError("processMessage response is missing")

            stdout = context.promise.result["processMessage"]["stdout"]
            matches = re.findall(r"<finalAnswer>(.*?)</finalAnswer>", stdout, re.DOTALL)
            if len(matches) > 0:
                context.status = RunnableStatus.SUCCESS
                context.response = ChainResponse(finalAnswer=matches[0],history=context.data["runtime"]["history"])
                return context
            
            matches = re.findall(r"<function>(.*?)</function>", stdout, re.DOTALL)
            if len(matches) > 0:
                fnCall = json.loads(matches[0])
                fnDefines = [fn for fn in context.request.functions if fn.name == fnCall["name"]]
                if len(fnDefines) == 0:
                    context.status = RunnableStatus.ERROR
                    context.errorMessage = f"Function:{fnCall['name']} not found"
                    return context
                
                fnDefine = fnDefines[0]

                if fnDefine.type == ChainFunctionType.TOOL:
                    runnableCode = "TOOL"
                elif fnDefine.type == ChainFunctionType.AGENT:
                    runnableCode = "AGENT"

                context.promise.resolve["functionCall"] = {
                    "runnableCode": runnableCode,
                    "toolCode": fnDefine.name,
                    "toolVersion": fnDefine.version,
                    "inputs": self.fnInputMerge(fnDefine, fnCall["input"]),
                }
                context.data["runtime"]["nextStep"] = "functionCallEnd"
                return context
            
            context.status = RunnableStatus.ERROR
            context.errorMessage = "finalAnswer or function not found"
            return context

        elif context.data["runtime"]["nextStep"] == "functionCallEnd":

            if context.promise.result["functionCall"] is None:
                raise ValueError("functionCall response is missing")
            
            context.promise.resolve["processMessage"] = {
                "runnableCode": "PYTHON",
                "data": {
                    "function": json.dumps(context.promise.result["functionCall"]["outputs"]),
                    "completion": context.data["runtime"]["lastCompletion"],
                },
                "run": context.request.onNext + "\n" + self.nextPostScript,
            }

            context.data["runtime"]["nextStep"] = "llmStart"
            return context

        elif context.data["runtime"]["nextStep"] == "llmStart":

            if context.promise.result["processMessage"] is None:
                raise ValueError("processMessage response is missing")

            stdout = context.promise.result["processMessage"]["stdout"]
            matches = re.findall(r"<message>(.*?)</message>", stdout, re.DOTALL)
            if len(matches) > 0:
                context.promise.resolve["llm"] = {
                    "runnableCode": "LLM",
                    "setting": context.request.llm.model_dump(),
                    "systemPrompt": context.data["runtime"]["systemPrompt"],
                    "userPrompt": matches[0],
                    "history": context.data["runtime"]["history"],
                }
                context.data["runtime"]["nextStep"] = "llmEnd"
                return context
            
            context.status = RunnableStatus.ERROR
            context.errorMessage = "message not found"
            return context 
        
        return context   
            






import os
from typing import Dict, List
from abc import ABC, abstractmethod

class RunnableFileStore(ABC):
    @abstractmethod
    def saveFile(self, filePath, content):
        pass

    @abstractmethod
    def readFile(self, filePath) -> str:
        pass

class RunnableDatabaseStore(ABC):
    @abstractmethod
    def queryRows(self, table:str, where:str) -> List[Dict]:
        pass

    @abstractmethod
    def insertRows(self, table:str, data:List[Dict]) -> str:
        pass


class RunnableLocalFileStore(RunnableFileStore):
    def __init__(self, path):
        self.path = path

    def saveFile(self, filePath, content):
        fullPath = os.path.join(self.path, filePath)
        directory = os.path.dirname(fullPath)
        
        # 检查目录是否存在，如果不存在则创建
        if not os.path.exists(directory):
            os.makedirs(directory)
        
        # 写入文件
        with open(fullPath, 'w') as file:
            file.write(content)

    def readFile(self, filePath) -> str:
        fullPath = os.path.join(self.path, filePath)
        
        # 检查文件是否存在
        if not os.path.exists(fullPath):
            raise FileNotFoundError(f"File {fullPath} does not exist")
        
        # 读取文件内容
        with open(fullPath, 'r') as file:
            return file.read()


from setuptools import setup, find_packages

setup(
    name='runnable_hub',
    version='0.3',
    packages=find_packages(),
    install_requires=[
        "pydantic",
        "aiohttp",
        "jinja2"
    ],
    author='jiongen.zje',
    author_email='jiongen.zje@alibaba-inc.com',
    description='Runnable Hub is a middleware for large models, enabling management and execution of complex inference tasks with a flexible architecture and support for multiple task types.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/alibaba/sreworks-ext/tree/main/runnable-hub',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.12',
)



from runnable_hub import RunnableWorker, RunnableContext, RunnableStatus
from runnable_hub.store import RunnableFileStore, RunnableDatabaseStore
from .request.toolRequest import ToolRequest
from .request.toolDefine import ToolDefine, ToolType
from .response import ToolResponse
from typing import Dict
from datetime import datetime
import json

class Worker(RunnableWorker):

    runnableCode = "TOOL"
    Request = ToolRequest
    Response = ToolResponse

    def __init__(self, store: RunnableFileStore|RunnableDatabaseStore):
        self.store = store

    def addTool(self, toolDefine: ToolDefine):
        if isinstance(self.store, RunnableFileStore):
            filePath = f"{toolDefine.toolCode}/{toolDefine.toolVersion}/define.json"
            self.store.saveFile(filePath, toolDefine.model_dump_json())
        else:
            raise Exception("Not supported")

    def readTool(self, toolCode:str, toolVersion:str) -> ToolDefine:
        if isinstance(self.store, RunnableFileStore):
            filePath = f"{toolCode}/{toolVersion}/define.json"
            return ToolDefine.model_validate_json(self.store.readFile(filePath))
        else:
            raise Exception("Not supported")

    async def onNext(self, context: RunnableContext[ToolRequest, ToolResponse]) -> RunnableContext:
        toolDefine = self.readTool(context.request.toolCode, context.request.toolVersion)
        if context.data.get("sendProcessRequest") is None:
            if toolDefine.toolType == ToolType.API:
                processStepName = "api"
            else:
                raise Exception("Not supported")

            processRequest = {
                "runnableCode": "PROCESS",
                "outputs": "${{ jobs.call.outputs }}",
                "jobs":{
                    "call": {
                        "outputs": "${{ steps.result.outputs }}",
                        "steps":[{
                            "id": "setting",
                            "jinja":{
                                "data": {
                                    "inputs": context.request.inputs
                                },
                                "outputLoads": "JSON",
                                "template": json.dumps(toolDefine.setting),
                            }
                        },{
                            "id": "call",
                            processStepName: "${{ steps.setting.outputs }}"
                        },{
                            "id": "result",
                            "jinja":{
                                "data": {
                                    "result": "${{ steps.call.outputs }}"
                                },
                                "outputLoads": toolDefine.outputsLoads,
                                "template": toolDefine.outputTemplate
                            }
                        }]
                    }
                }
            }
            context.promise.resolve["todo"] = processRequest
            context.data["sendProcessRequest"] = datetime.now()
        else:
            if context.promise.result.get("todo") is None:
                raise Exception("Process request not finish")
            
            result = context.promise.result["todo"]

            context.response = ToolResponse(
                success=True,
                outputs=result["outputs"]
            )
            context.status = RunnableStatus.SUCCESS

        return context







from runnable_hub import RunnableRequest, RunnableOutputLoads
from .apiHttpMethod import ApiHttpMethod
from typing import Dict

class ApiRequest(RunnableRequest):
    runnableCode: str = "API"
    url: str
    method: ApiHttpMethod = ApiHttpMethod.GET
    outputLoads: RunnableOutputLoads = RunnableOutputLoads.TEXT
    headers: Dict = {}
    params: Dict = {}
    payloads: Dict = {}



from runnable_hub import RunnableResponse
from typing import Dict, List, Optional
from .llmUsage import LlmUsage
from ..request.llmMessage import LlmMessage

class LlmResponse(RunnableResponse):
    runnableCode: str = "LLM"
    usage: LlmUsage
    content: str
    messages: List[LlmMessage]